/*
 * The Copyright of this work is vested in the National Electricity Market
 * Management Company Limited and the software is issued in confidence
 * for the purpose only for which it is supplied. It must not be
 * reproduced in whole or in part or used for tendering or manufacturing
 * purposes except under an agreement or with the consent in writing of
 * the National Electricity Market Management Company Limited and then
 * only on the condition that this notice is included in any such
 * reproduction.  No information as to the contents or subject matter of
 * this software or any part thereof arising directly or indirectly there
 * from shall be given orally or in writing or communicated in any manner
 * whatsoever to any third party being an individual firm or company or
 * any employee thereof without the prior consent in writing of the National
 * Electricity Market Management Company Limited.
 *
 * Copyright the National Electricity Market Management Company Limited.  2003
 * Mike Muir
 */

package com.nemmco.participantbatcher.ftp;

import com.enterprisedt.net.ftp.FTPClient;
import com.enterprisedt.net.ftp.FTPConnectMode;
import com.enterprisedt.net.ftp.FTPException;
import com.enterprisedt.net.ftp.FTPTransferType;

import com.nemmco.participantbatcher.util.Constants;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;

/**
 * This class is the FTP process implementation based on Enterprise Distributed Technologies Ltd
 * third party library.
 *
 * http://www.enterprisedt.com
 *
 * Their Copyright Statement see their software
 *
 *
 * <p>
 * author Mike Muir
 * version 1.4.4 - 24-Apr-2006 - Creation as part of testing FTP performance via NAT, set port range
 * version 1.5.0 - 08-Aug-2006 - added  webDAV
 * version 2.0.1 - 15-Nov-2010 - Peer code review
 * version 2.0.2 - 19-Nov-2010 - Peer code review
 * @author Deepak Gupta
 * @version 46.93 - 6-Mar-2018 - DFCT0021839
 *
 */

public class FTPEDTProcesser implements ComProcess {

   private FTPClient ftp = null;
   private String cwd = "";

   /**
    * Constructor
    */
   public FTPEDTProcesser() {
      ftp = new FTPClient();
   }

   /**
    * Connect to the FTP server.
    * <p>
    * @param serverName String FTP server name
    * @exception ComException If an error occurs while connecting.
    */
   public void connect(String serverName) throws ComException {
      try {
         ftp.setRemoteHost(serverName);
         ftp.connect();
      } catch (FTPException e) {
         throw new ComException("Exception occured when connecting to FTP server - " + serverName + " code " + e.getReplyCode(), e);
      } catch (Exception e) {
         throw new ComException("Exception occured when connecting to FTP server - " + serverName, e);
      }
   }

   /**
    * Set the port working range in active ftp mode.
    * <p>
    * @param lowLimit int
    * @param highLimit int
    * @exception ComException If an error occurs while setting range
    */
   public void setActivePortRange(int lowLimit, int highLimit) throws ComException {
      try {
         ftp.setActivePortRange(lowLimit, highLimit);
      } catch (Exception e) {
         throw new ComException("Exception occured when setting port range", e);
      }
   }

   /**
    * set the domain for webDAV
    * @param domain String
    */
   public void setDomain(String domain) throws ComException {
      // do nothing for FTP
   }

   /**
    * Closes the connection to the FTP server.
    * <p>
    * @exception ComException If an error occurs while disconnecting.
    */
   public void disconnect() throws ComException {
      try {
         if (ftp.connected()) {
            ftp.quit();
         }
      } catch (FTPException e) {
         throw new ComException("Exception occured when disconnecting from FTP server code " + e.getReplyCode(), e);
      } catch (Exception e) {
         throw new ComException("Exception occured when disconnecting from FTP server ", e);
      }
   }

   /**
    * Test if the FTP connection is still connected.
    * <p>
    * @return TRUE if the FTP connection is still connected
    * @exception ComException If an error occurs when checking connectivity.
    */
   public boolean isConnected() throws ComException {
      try {
         return ftp.connected();
      } catch (Exception e) {
         throw new ComException("Exception occured when checking connectivity", e);
      }
   }

   /**
    * Login to the FTP server using the provided username and password.
    *
    * @param userName The username to login under.
    * @param password The password to use.
    * @return True if successfully completed, false if not.
    * @exception ComException  If an error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean login(String userName, String password) throws ComException {
      //boolean returnCode = false;
      try {
         ftp.login(userName, password);
			Constants c = Constants.getConstant();
         if (Constants.getConstant().getFtpPassiveMode()) { // V1.3.3 go into passive mode {
            ftp.setConnectMode(FTPConnectMode.PASV);
         } else {
            ftp.setConnectMode(FTPConnectMode.ACTIVE);
         }
			int bufferSize = c.getBufferSize();
			if (bufferSize > 0) {
				ftp.setTransferBufferSize(bufferSize);
			}
         cwd = "/";
         return isConnected();
      } catch (FTPException e) {
         throw new ComException("Exception occured when logging into FTP server code " + e.getReplyCode(), e);
      } catch (Exception e) {
         throw new ComException("Exception occured when logging into FTP server", e);
      }
   }

   /**
    * Logout of the FTP server by sending the QUIT command.
    * <p>
    * @return True if successfully completed, false if not.
    * @exception ComException  If an error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean logout() throws ComException {
      try {
         ftp.quit();
         return true;
      } catch (FTPException e) {
         throw new ComException("Exception occured when logging out code " + e.getReplyCode(), e);
      } catch (Exception e) {
         throw new ComException("Exception occured when logging out", e);
      }
   }

   /**
    * Test if the reply from FTP server is positive.
    * <p>
    * @return boolean TRUE if positive reply, FALSE if negative reply
    * @exception ComException If an error occurs while disconnecting.
    */
   public boolean isPositiveReply() throws ComException {
      try {
         ftp.noOperation();
         return true;
      } catch (FTPException e) {
         throw new ComException("Exception occured when getting reply code from FTP server code " + e.getReplyCode(), e);
      } catch (Exception e) {
         throw new ComException("Exception occured when getting reply code from FTP server", e);
      }
   }

   /**
    * Sets the timeout in milliseconds to use when processing (downloading
    * and uploading) files.  This timeout should be set immediately after
    * opening the FTP connection. and the value should be able to be set
    * in the configuration file.
    * <p>
    * @param  timeout The default timeout in milliseconds that is used when
    *         processing a file.
    */
   public void setDataTimeout(int timeout) {
      try {
         ftp.setTimeout(timeout);
      } catch (Exception e) {
         throw new ComException("Exception occured setting timeout value", e);
      }
   }

   /**
    * Change the current working directory of the FTP session.
    *
    * @param pathName  The new current working directory.
    * @return True if successfully completed, false if not.
    * @exception ComException  If an error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean changeWorkingDirectory(String pathName) throws ComException {
      try {
         ftp.chdir(pathName);
         cwd = pathName;
         return true;
      } catch (FTPException e) {
         throw new ComException("Exception occured when changing working directory to " + pathName + " code " + e.getReplyCode(), e);
      } catch (Exception e) {
         throw new ComException("Exception occured when changing working directory to " + pathName, e);
      }
   }

   /**
    * Sets the file type to be transferred.
    * This should be one of ASCII, BINARY for Participant Batcher
    * The file type only needs to be set when you want to change the type.
    * After changing it, the new type stays in effect until you change it
    * again.  The default file type is ASCII
    * <p>
    * @param      fileType either ASCII or BINARY indcating the type of file.
    * @return     True if successfully completed, false if not.
    * @exception  ComException  If an error occurs while either sending a
    *             command to the server or receiving a reply from the server.
    */
   public boolean setFileType(String fileType) throws ComException {
      try {
         if (Constants.BINARY_FILE_TYPE.equalsIgnoreCase(fileType)) {
            ftp.setType(FTPTransferType.BINARY);
         } else {
            ftp.setType(FTPTransferType.ASCII);
         }
         return true;
      } catch (FTPException e) {
         throw new ComException("Exception occured when setting file type code " + e.getReplyCode(), e);
      } catch (Exception e) {
         throw new ComException("Exception occured when setting file type", e);
      }
   }

   /**
    * Retrieves a named file from the server and writes it to the given file.
    * <p>
    * @param remote  The name of the remote file.
    * @param local   The local file to which to write.
    * @return True if successfully completed, false if not.
    * @exception ComException  If an error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean retrieveFile(String remote, File local) throws ComException {
      try {
         FileOutputStream fos = new FileOutputStream(local);
         try {
            ftp.get(fos, remote);
         } finally {
            fos.close();
         }
         return true;
      } catch (FTPException e) {
         if (local.delete()) { // v1.4.4 delete the local version which could be a zero length file
            throw new ComException("Exception occured when retrieving file " + remote + " from FTP server code " + e.getReplyCode(), e);
         } else {
            throw new ComException("Exception occured when retrieving file " + remote + " from FTP server code " + e.getReplyCode() + ",local delete also failed", e);
         }
      } catch (Exception e) {
         if (local.delete()) { // v1.4.4 delete the local version which could be a zero length file
            throw new ComException("Exception occured when retrieving file " + remote + " from FTP server", e);
         } else {
            throw new ComException("Exception occured when retrieving file " + remote + " from FTP server, local delete also failed", e);
         }
      }
   }

   /**
    * Stores a file on the server using the given name and taking input
    * from the given file.
    * <p>
    * @param remote  The name to give the remote file.
    * @param local   The local file from which to read.
    * @return True if successfully completed, false if not.
    * @exception ComException  If an I/O error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean storeFile(String remote, File local) throws ComException {
      try {
         FileInputStream fis = new FileInputStream(local);
         try {
            ftp.put(fis, remote);
         } finally {
            fis.close();
         }
         return true;
      } catch (FTPException e) {
         throw new ComException("Exception occured when storing file " + remote + " to FTP server code " + e.getReplyCode(), e);
      } catch (Exception e) {
         throw new ComException("Exception occured when storing file " + remote + " to FTP server", e);
      }
   }

   /**
    * Renames a remote file.
    * <p>
    * @param from  The name of the remote file to rename.
    * @param to    The new name of the remote file.
    * @return True if successfully completed, false if not.
    * @exception ComException  If an I/O error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean rename(String from, String to) throws ComException {
      try {
         ftp.rename(from, to);
         return true;
      } catch (FTPException e) {
         throw new ComException("Exception occured when renaming file " + from + " to " + to + " code " + e.getReplyCode(), e);
      } catch (Exception e) {
         throw new ComException("Exception occured when renaming file " + from + " to " + to, e);
      }
   }

   /**
    * Abort a transfer in progress.
    * <p>
    * @return True if successfully completed, false if not.
    * @exception ComException  If an I/O error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean abort() throws ComException {
      try {
         //ftp.abort();  this should not be used as it may leave partial files
         return true;
      } catch (Exception e) {
         throw new ComException("Exception occured when aborting process", e);
      }
   }

   /***
    * Deletes a file on the FTP server.
    * <p>
    * @param fileName   The fileName of the file to be deleted.
    * @return True if successfully completed, false if not.
    * @exception ComException  If an I/O error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    ***/
   public boolean deleteFile(String fileName) throws ComException {
      try {
         ftp.delete(fileName);
         return true;
      } catch (FTPException e) {
         throw new ComException("Exception occured when deleting " + fileName + " code " + e.getReplyCode(), e);
      } catch (Exception e) {
         throw new ComException("Exception occured when deleting " + fileName, e);
      }
   }

   /***
    * Obtain a list of filenames in a directory.
    * If the given pathname contains no files,  a zero length array or null is
    * returned. If the directory is not empty, an array of filenames in the
    * directory is returned. If the pathname corresponds to a file, only that
    * file will be listed.
    *
    * @param pathName  The directory or file to list.
    *                  NULL or empty string stands for current working directory
    * @return The list of filenames contained in the given path.  null if
    *     the list could not be obtained.  If there are no filenames in
    *     the directory, a zero-length array is returned.
    * @exception ComException  If an error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    ***/
   public String[] listNames(String pathName) throws ComException {
      try {
         ftp.chdir(pathName);
         String[] fileList = ftp.dir();
         ftp.chdir(cwd);
         return fileList;
      } catch (FTPException e) {
         throw new ComException("Exception occured when getting file list for " + pathName + " code " + e.getReplyCode(), e);
      } catch (Exception e) {
         throw new ComException("Exception occured when getting file list for " + pathName, e);
      }
   }

}
