/*
 * The Copyright of this work is vested in the National Electricity Market
 * Management Company Limited and the software is issued in confidence
 * for the purpose only for which it is supplied. It must not be
 * reproduced in whole or in part or used for tendering or manufacturing
 * purposes except under an agreement or with the consent in writing of
 * the National Electricity Market Management Company Limited and then
 * only on the condition that this notice is included in any such
 * reproduction.  No information as to the contents or subject matter of
 * this software or any part thereof arising directly or indirectly there
 * from shall be given orally or in writing or communicated in any manner
 * whatsoever to any third party being an individual firm or company or
 * any employee thereof without the prior consent in writing of the National
 * Electricity Market Management Company Limited.
 *
 * Copyright the National Electricity Market Management Company Limited.  2003
 * Mike Muir
 */

package com.nemmco.participantbatcher.ftp;

import com.nemmco.participantbatcher.util.Constants;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;

import org.apache.commons.net.ftp.FTP;
import org.apache.commons.net.ftp.FTPClient;
import org.apache.commons.net.ftp.FTPReply;

/**
 * This class is the FTP process implementation based on Apache commons-net-1.0.0
 * third party library.
 *
 * http://jakarta.apache.org/commons/net/apidocs/org/apache/commons/net/ftp/FTPClient.html
 *
 * <p>
 * author Kevin Jin
 * version 1.0 - 21 Aug 2003 - Creation
 * version 1.3.1 - 15-Nov-2004 Put the passive mode after the connect
 * version 1.3.3 - 16-Feb-2005 Put the FTPMODE as a configurable item
 * version 1.4.0 - 28-Feb-2005 Added share as alternative to FTP
 * version 1.4.4 - 22-Apr-2006 Free file connection on FTP error and delete local file if FTP failure has occured
 * version 1.5.0 - 08-Aug-2006 Addeb webDAV
 * version 2.0.2 - 19-Nov-2010 Peer review work
 * @author Deepak Gupta
 * @version 46.93 - 6-Mar-2018 - DFCT0021839
 */

public class FTPProcesser implements ComProcess {

   private FTPClient ftp = null;

   /**
    * Constructor
    */
   public FTPProcesser() {
      ftp = new FTPClient();
      //ftp.enterLocalPassiveMode();
   }

   /**
    * Connect to the FTP server.
    * <p>
    * @param serverName FTP server name  String
    * @exception ComException If an error occurs while connecting.
    */
   public void connect(String serverName) throws ComException {
      try {
         ftp.connect(serverName);
         if (Constants.getConstant().getFtpPassiveMode()) { // V1.3.3 go into passive mode
            ftp.enterLocalPassiveMode();
         } else {
            ftp.enterLocalActiveMode();
         }
      } catch (Exception e) {
         throw new ComException("Exception occured when connecting to FTP server - " + serverName, e);
      }
   }

   /**
    * Set the port working range in active ftp mode.  Dosent work in FTPProcesser.
    * <p>
    * @param lowLimit int
    * @param highLimit int
    */
   public void setActivePortRange(int lowLimit, int highLimit) throws ComException {}

   /**
    * set the domain for webDAV
    * @param domain String
    */
   public void setDomain(String domain) throws ComException {
      // do nothing for FTP
   }

   /**
    * Closes the connection to the FTP server.
    * <p>
    * @exception ComException If an error occurs while disconnecting.
    */
   public void disconnect() throws ComException {
      try {
         ftp.disconnect();
      } catch (Exception e) {
         throw new ComException("Exception occured when disconnecting from FTP server", e);
      }
   }

   /**
    * Test if the FTP connection is still connected.
    * <p>
    * @return TRUE if the FTP connection is still connected
    * @exception ComException If an error occurs when checking connectivity.
    */
   public boolean isConnected() throws ComException {
      try {
         return ftp.isConnected();
      } catch (Exception e) {
         throw new ComException("Exception occured when checking connectivity", e);
      }
   }

   /**
    * Login to the FTP server using the provided username and password.
    *
    * @param userName The username to login under.
    * @param password The password to use.
    * @return True if successfully completed, false if not.
    * @exception ComException  If an error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean login(String userName, String password) throws ComException {
      try {
			boolean loggedIn = ftp.login(userName, password);
			if (loggedIn) {
				int bufferSize = Constants.getConstant().getBufferSize();
				if (bufferSize > 0) {
					ftp.setBufferSize(bufferSize);
				}
			}
			return loggedIn;
      } catch (Exception e) {
         throw new ComException("Exception occured when logging into FTP server", e);
      }
   }

   /**
    * Logout of the FTP server by sending the QUIT command.
    * <p>
    * @return True if successfully completed, false if not.
    * @exception ComException  If an error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean logout() throws ComException {
      try {
         return ftp.logout();
      } catch (Exception e) {
         throw new ComException("Exception occured when logging out", e);
      }
   }

   /**
    * Test if the reply from FTP server is positive.
    * <p>
    * @return boolean TRUE if positive reply, FALSE if negative reply
    * @exception ComException If an error occurs while disconnecting.
    */
   public boolean isPositiveReply() throws ComException {
      try {
         int reply;
         reply = ftp.getReplyCode();
         return FTPReply.isPositiveCompletion(reply);
      } catch (Exception e) {
         throw new ComException("Exception occured when getting reply code from FTP server", e);
      }
   }

   /**
    * Sets the timeout in milliseconds to use when processing (downloading
    * and uploading) files.  This timeout should be set immediately after
    * opening the FTP connection. and the value should be able to be set
    * in the configuration file.
    * <p>
    * @param  timeout The default timeout in milliseconds that is used when
    *         processing a file.
    */
   public void setDataTimeout(int timeout) {
      ftp.setDataTimeout(timeout);
   }

   /**
    * Change the current working directory of the FTP session.
    * <p>
    * @param pathName  The new current working directory.
    * @return True if successfully completed, false if not.
    * @exception ComException  If an error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean changeWorkingDirectory(String pathName) throws ComException {
      try {
         return ftp.changeWorkingDirectory(pathName);
      } catch (Exception e) {
         throw new ComException("Exception occured when changing working directory to " + pathName, e);
      }
   }

   /**
    * Sets the file type to be transferred.
    * This should be one of ASCII, BINARY for Participant Batcher
    * The file type only needs to be set when you want to change the type.
    * After changing it, the new type stays in effect until you change it
    * again.  The default file type is ASCII
    * <p>
    * @param      fileType either ASCII or BINARY indcating the type of file.
    * @return     True if successfully completed, false if not.
    * @exception  ComException  If an error occurs while either sending a
    *             command to the server or receiving a reply from the server.
    */
   public boolean setFileType(String fileType) throws ComException {
      try {
         boolean result = false;
         if (Constants.BINARY_FILE_TYPE.equalsIgnoreCase(fileType)) {
            result = ftp.setFileType(FTP.BINARY_FILE_TYPE);
         } else {
            result = ftp.setFileType(FTP.ASCII_FILE_TYPE);
         }
         return result;
      } catch (Exception e) {
         throw new ComException("Exception occured when setting file type", e);
      }
   }

   /**
    * Retrieves a named file from the server and writes it to the given file.
    * <p>
    * @param remote  The name of the remote file.
    * @param local   The local file to which to write.
    * @return True if successfully completed, false if not.
    * @exception ComException  If an error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean retrieveFile(String remote, File local) throws ComException {
      boolean result = false;
      try {
         FileOutputStream fos = new FileOutputStream(local);
         try {
            result = ftp.retrieveFile(remote, fos);
         } finally {
            fos.close();
         }
      } catch (Exception e) {
         if (local.delete()) { // v1.4.4 delete the local version which could be a zero length file
            throw new ComException("Exception occured when retrieving file " + remote + " from FTP server, local file cleared", e);
         } else {
            throw new ComException("Exception occured when retrieving file " + remote + " from FTP server, local file not cleared", e);
         }
      }
      return result;
   }

   /**
    * Stores a file on the server using the given name and taking input
    * from the given file.
    * <p>
    * @param remote  The name to give the remote file.
    * @param local   The local file from which to read.
    * @return True if successfully completed, false if not.
    * @exception ComException  If an I/O error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean storeFile(String remote, File local) throws ComException {
      try {
         boolean result = false;
         FileInputStream fis = new FileInputStream(local);
         try {
            result = ftp.storeFile(remote, fis);
         } finally {
            fis.close();
         }
         return result;
      } catch (Exception e) {
         throw new ComException("Exception occured when storing file " + remote + " to FTP server", e);
      }
   }

   /**
    * Renames a remote file.
    * <p>
    * @param from  The name of the remote file to rename.
    * @param to    The new name of the remote file.
    * @return True if successfully completed, false if not.
    * @exception ComException  If an I/O error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean rename(String from, String to) throws ComException {
      try {
         return ftp.rename(from, to);
      } catch (Exception e) {
         throw new ComException("Exception occured when renaming file " + from + " to " + to, e);
      }
   }

   /**
    * Abort a transfer in progress.
    * <p>
    * @return True if successfully completed, false if not.
    * @exception ComException  If an I/O error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    */
   public boolean abort() throws ComException {
      try {
         return ftp.abort();
      } catch (Exception e) {
         throw new ComException("Exception occured when aborting process", e);
      }
   }

   /***
    * Deletes a file on the FTP server.
    * <p>
    * @param fileName   The fileName of the file to be deleted.
    * @return True if successfully completed, false if not.
    * @exception ComException  If an I/O error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    ***/
   public boolean deleteFile(String fileName) throws ComException {
      try {
         return ftp.deleteFile(fileName);
      } catch (Exception e) {
         throw new ComException("Exception occured when deleting " + fileName, e);
      }
   }

   /***
    * Obtain a list of filenames in a directory.
    * If the given pathname contains no files,  a zero length array or null is
    * returned. If the directory is not empty, an array of filenames in the
    * directory is returned. If the pathname corresponds to a file, only that
    * file will be listed.
    *
    * @param pathName  The directory or file to list.
    *                  NULL or empty string stands for current working directory
    * @return The list of filenames contained in the given path.  null if
    *     the list could not be obtained.  If there are no filenames in
    *     the directory, a zero-length array is returned.
    * @exception ComException  If an error occurs while either sending a
    *      command to the server or receiving a reply from the server.
    ***/
   public String[] listNames(String pathName) throws ComException {
      try {
         String[] fileList = ftp.listNames(pathName);
         return fileList;
      } catch (Exception e) {
         throw new ComException("Exception occured when getting file list for " + pathName, e);
      }
   }

}
