/*
 * The Copyright of this work is vested in the National Electricity Market
 * Management Company Limited and the software is issued in confidence
 * for the purpose only for which it is supplied. It must not be
 * reproduced in whole or in part or used for tendering or manufacturing
 * purposes except under an agreement or with the consent in writing of
 * the National Electricity Market Management Company Limited and then
 * only on the condition that this notice is included in any such
 * reproduction.  No information as to the contents or subject matter of
 * this software or any part thereof arising directly or indirectly there
 * from shall be given orally or in writing or communicated in any manner
 * whatsoever to any third party being an individual firm or company or
 * any employee thereof without the prior consent in writing of the National
 * Electricity Market Management Company Limited.
 *
 * Copyright the National Electricity Market Management Company Limited.  2003
 * Mike Muir
 * 28-Aug-2012     Oleg Pevzner       7984 Java upgrade from 6 to 7
 */

package com.nemmco.participantbatcher.process;

import com.nemmco.participantbatcher.ftp.ComProcess;
import com.nemmco.participantbatcher.util.Constants;
import com.nemmco.participantbatcher.util.HttpDataStore;
import com.nemmco.participantbatcher.util.Util_Logger;
import com.nemmco.participantbatcher.xml.AckFileParser;
import com.nemmco.participantbatcher.xml.ZipFileInfo;
import com.nemmco.participantbatcher.xml.ZipFileValidator;

import java.io.File;

import org.apache.log4j.Logger;

/**
 * This class is used to control the process of User sends Zip files via FTP.
 *
 * @author Kevin Jin
 * version 1.0 - 28 Aug 2003 - Creation
 * version 1.1 - 4 Nov 2003 modified logging
 * version 1.3 - 19-May-2004 added stop file processing,
 *                20-Aug-2004 fixed the repetitive processing of ack files, event logging,
 *                stopbox processing for b2b files, log4j, update control file for resent files
 * version 1.3.1 - 22-Nov-2004 fixed null return for a dir on a non existant stopbox, high/low water mark logic
 * version 1.4.0 - 28-Feb-2005 Added share as alternative to FTP
 * version 1.4.2 - 13-01-2006 Fixed defect where the .ack file was downloaded many times
 *                  giving spurious warnings when attempts were made to delete the associated zip
 *                  This happens if pb is faster than the cleanack inbound .ack removal
 *                  - Added data to HttpDataStore
 * version 1.4.4 - 22-Apr-2006 When an FTP exception occurs the .ack may have been uploaded and the .zip deleted
 *                              at the NEMMCO end so that .sent gets renamed to .zip in next round - so check for
 *                              ack* file before renaming of .sent back to .zip
 *                  30-Apr-2006 Make sure the transfer file type is specified before all FTP file transfers
 * version 1.5.0 - 23-June-2006 Changed stopping for Instance Batcher functionality
 * version 1.5.1 - 1-Nov-2006 Made BatcherFileFilter generic to handle a file type or a set of file types, replaces AckFileFilter
 * version 1.5.2 - 8-Mar-2007 Fix sending of B2B to send any file that is schema invalid or zip invalid (so that it gets -ve acked)
 * version 1.5.3 - 30-Jun-2007 Change B2B stop file processing to work with participant names that are subsets of other names eg DELTA DELTAX
 * version 1.5.4 - 8-Aug-2007 InstanceBatcher2 return work done
 * version 2.0.1 - 15-Nov-2010 Fixed bug with the renaming of .sent files back to .zip files, Peer review work
 * version 2.0.2 - 19-Nov-2010 Peer review work
 * @version 2.0.3 - 22 - Feb - 2011 Fix null pliACKFileExistList null pointer exception - test before using
 *
 */

public class SendProcess {

   // Directories, for easy coding
   private String pnInbox = "";
   private String pnOutbox = "";
   private String plInbox = "";
   //private String plOutbox = "";
   private int zipsSent = 0;
   private int zipsDeleted = 0;
   private ComProcess processer = null;
   private static Logger logger = Logger.getLogger(SendProcess.class);
   private HttpDataStore hds;
   private FileManager fm = null;
   private AckFileParser afp = null;
   private ZipFileValidator zfv = null;

   public SendProcess() {
      this.pnInbox = Constants.getConstant().getPNInboxDir();
      this.pnOutbox = Constants.getConstant().getPNOutboxDir();
      this.plInbox = Constants.getConstant().getPLInboxDir();
      //this.plOutbox = Constants.getConstant().getPLOutboxDir();
      this.hds = HttpDataStore.getHttpDataStore();
   }

   /**
    * This method is used to control the sending of zip files to the hub.
    *
    * @param processerParam ComProcess the main communication processer eg FTP
    * @param pniRawFileList String[] file list in the participant NEMMCO inbox
    * @param pnoRawFileList String[] file list in the participant NEMMCO outbox
    */
   public int process(ComProcess processerParam, String[] pniRawFileList, String[] pnoRawFileList) {

      processer = processerParam;

      Util_Logger.debug(logger, "Start Sending Process");
      long startTime = System.currentTimeMillis();

      fm = new FileManager();
      zfv = new ZipFileValidator();
      afp = new AckFileParser();

      try {

         zipsSent = 0;
         zipsDeleted = 0;
         processScan(pniRawFileList, pnoRawFileList);

         // log informative message if progress made
         long timeTaken = System.currentTimeMillis() - startTime;
         if (zipsDeleted + zipsSent > 0) {
            Util_Logger.info(logger, "End Send. " + zipsDeleted + " zips deleted, " + zipsSent + " zips sent, " + timeTaken + " msec");
         } else {
            Util_Logger.debug(logger, "End Send. " + zipsDeleted + " zips deleted, " + zipsSent + " zips sent, " + timeTaken + " msec");
         }
      } finally {
         hds.setZipsSent(zipsSent);
         hds.setAcksRecieved(zipsDeleted);
      }
      return zipsSent + zipsDeleted;
   }

   private void processScan(String[] pniRawFileList, String[] pnoRawFileList) {

      /*********************************************************************
         * For each .SENDING file in PLInbox, rename it back to .ZIP file
         * *******************************************************************/
      String[] pliSENDINGFileList = fm.scanDir(plInbox, new BatcherFileFilter(".sending"));
      if ((pliSENDINGFileList != null) && (pliSENDINGFileList.length > 0)) {
         for (int c = 0; c < pliSENDINGFileList.length; c++) {
            String pliSENDINGFile = pliSENDINGFileList[c];
            String pliZIPFile = fm.replaceFileName(pliSENDINGFile, "zip");
            fm.renameFile(plInbox, pliSENDINGFile, pliZIPFile);
         }
      }

      /*********************************************************************
         * For each .SENT file in PLInbox, which don't have a correspond .ZIP
         * file in PNInbox, rename it back to .ZIP file
         * *******************************************************************/

      // Get the .ZIP file list from the PNInbox
      Object[] pniZIPFileList = fm.scanFTPFileList(pniRawFileList, Constants.getConstant().getFileTypes(), ".zip");

      // Get the .ack file list from the PNOutbox
      Object[] pnoACKFileList = fm.scanFTPFileList(pnoRawFileList, Constants.getConstant().getFileTypes(), ".ack");

      // Get the .sent file list from plInbox
      String[] pliSENDFileList = fm.scanDir(plInbox, new BatcherFileFilter(".sent"));

      // Get the .ack* file list from plInbox v 1.4.4
      String[] pliAckAllFileExistList = fm.scanDir(plInbox, new BatcherFileFilter(new String[] { ".ack", ".ack_done", ".ack_resend", ".ack_nack" }));

      if ((pliSENDFileList != null) && (pliSENDFileList.length > 0)) {
         for (int c1 = 0; c1 < pliSENDFileList.length; c1++) {

            String pliSENDFile = pliSENDFileList[c1];
            String pniZIPFile = fm.replaceFileName(pliSENDFile, "zip");
            String pnoACKFile = fm.replaceFileName(pliSENDFile, "ack");

            boolean hasMatching = false;

            if ((pniZIPFileList != null) && (pniZIPFileList.length > 0)) {
               for (int c2 = 0; c2 < pniZIPFileList.length; c2++) {
                  if (((String)pniZIPFileList[c2]).indexOf(pniZIPFile) >= 0) {
                     hasMatching = true;
                     break;
                  }
               }
            }

            // v 1.4.4 check that no local .ack* exist before cleanup of .sent
            if ((!hasMatching) && (pliAckAllFileExistList != null) && (pliAckAllFileExistList.length > 0)) {

               String pliACKFile = fm.replaceFileName(pliSENDFile, "ack");
               for (int c12 = 0; c12 < pliAckAllFileExistList.length; c12++) {
                  if ((pliAckAllFileExistList[c12]).indexOf(pliACKFile) >= 0) {
                     hasMatching = true;
                     //Util_Logger.debug(logger, "Special " +(String)pliAckAllFileExistList[c12]+" send= "+ pliSENDFile + " ack= " + pliACKFile + " in " + plInbox);
                     break;
                  }
               }
            }

            // v2.0.1 Make sure the ack is not in the PNOutbox
            if ((!hasMatching) && (pnoACKFileList != null) && (pnoACKFileList.length > 0)) {
               for (int c22 = 0; c22 < pnoACKFileList.length; c22++) {
                  if (((String)pnoACKFileList[c22]).indexOf(pnoACKFile) >= 0) {
                     hasMatching = true;
                     break;
                  }
               }
            }

            if (!hasMatching) {
               fm.renameFile(plInbox, pliSENDFile, pniZIPFile);
               Util_Logger.debug(logger, "Renamed " + pliSENDFile + " back to " + pniZIPFile + " in " + plInbox);
            }

         }
      }

      /******************************************************************
         * For each .ZIP file in PLInbox, upload them to PNInbox directory
         * *****************************************************************/

      // Get File List from PLInbox
      String[] pliZIPFileList = fm.scanDir(plInbox, new BatcherFileFilter(".zip"));
      if (pliZIPFileList != null) {
         hds.setPliZipFileCount(pliZIPFileList.length);
      }

      if ((pliZIPFileList != null) && (pliZIPFileList.length > 0)) {
         // First, test high and low water marks on PNInbox
         int pniFileCount = pniZIPFileList.length;
         if (pniFileCount <= Constants.getConstant().getInboxLowMark()) {
            // send only enough files to keep NEMMCO inbox at HWM if below the LWM
            int uploadFileCount = Constants.getConstant().getInboxHighMark() - pniFileCount;
            if (pliZIPFileList.length < uploadFileCount) {
               uploadFileCount = pliZIPFileList.length;
            }

            Util_Logger.debug(logger, uploadFileCount + " ZIP files will be uploaded");

            // start a counter so that we send only uploadFileCount files
            int c8 = 0;

            // Now let start to upload by scaning the file list
            for (int c3 = 0; c3 < pliZIPFileList.length; c3++) {

               // only send the number needed
               if (c8 < uploadFileCount) {

                  // next file to look at
                  String pliZIPFile = pliZIPFileList[c3];

                  // test to see if the b2b test indicates the file should be skipped
                  if (checkB2B(pliZIPFile)) {
                     continue;
                  }

                  // increment the upload counter
                  c8++;

                  String pliSENDINGFile = fm.replaceFileName(pliZIPFile, "sending");
                  // rename to .sending
                  if (fm.renameFile(plInbox, pliZIPFile, pliSENDINGFile)) {
                     // upload file
                     String pniTMPFile = pnInbox + "/" + fm.replaceFileName(pliZIPFile, "tmp");
                     String pniZIPFile = pnInbox + "/" + fm.replaceFileName(pliZIPFile, "zip");
                     File pliZIP = new File(plInbox + "/" + pliSENDINGFile);

                     // Start to upload
                     boolean uploadSucceed = false;
                     if (processer.setFileType(Constants.BINARY_FILE_TYPE)) {
                        if (processer.storeFile(pniTMPFile, pliZIP)) {
                           if (processer.rename(pniTMPFile, pniZIPFile)) {
                              uploadSucceed = true;
                           }
                        }
                     }

                     if (uploadSucceed) {
                        // rename local to .sent
                        String pliSENDFile = fm.replaceFileName(pliZIPFile, "sent");
                        fm.renameFile(plInbox, pliSENDINGFile, pliSENDFile);

                        Util_Logger.debug(logger, "Upload " + plInbox + "/" + pliZIPFile + " successfully");
                        zipsSent++;
                     } else {
                        // rename local .sending back to .zip
                        fm.renameFile(plInbox, pliSENDINGFile, pliZIPFile);

                        Util_Logger.debug(logger, "Upload " + plInbox + "/" + pliZIPFile + " failed");
                     }
                  }

                  Util_Logger.debug(logger, "Waiting for " + Constants.getConstant().getFileWaitTime());

                  //Thread t = new Thread();  //java 1.6
                  try {
                     Thread.sleep(Constants.getConstant().getFileWaitTime());
                  } catch (Exception e) {
                     ; // do nothing
                  }

               }

            }

            Util_Logger.debug(logger, uploadFileCount + " ZIP files may have been uploaded");

         }
      }

      /**********************************************************************
         * Download ACK file and delete related ZIP file
         **********************************************************************/

      String[] pliACKFileExistList = fm.scanDir(plInbox, new BatcherFileFilter(new String[] { ".ack", ".ack_done", ".ack_resend", ".ack_nack" }));

      if ((pnoACKFileList != null) && (pnoACKFileList.length > 0)) {

         Util_Logger.debug(logger, pnoACKFileList.length + " ACK files may be processed");

         for (int c4 = 0; c4 < pnoACKFileList.length; c4++) {
            String pnoACKFile = (String)pnoACKFileList[c4];

            boolean hasMatch = false;
            if ((pliACKFileExistList != null) && (pliACKFileExistList.length > 0)) {
               for (int c9 = 0; c9 < pliACKFileExistList.length; c9++) {
                  // V 1.4.2
                  if ((pliACKFileExistList[c9]).indexOf(pnoACKFile) >= 0) {
                     hasMatch = true;
                     break;
                  }
               }
            }

            // if NOT already downloaded in a previous cycle v1.4.4
            if (!hasMatch) {

               Util_Logger.debug(logger, "Start FTP process " + pnoACKFile);

               String pliACKFile = plInbox + "/" + pnoACKFile;
               if (processer.setFileType(Constants.ASCII_FILE_TYPE)) {
                  if (processer.retrieveFile(pnOutbox + "/" + pnoACKFile, new File(pliACKFile))) {
                     Util_Logger.debug(logger, "Download " + pnOutbox + "/" + pnoACKFile + " successfully");

                     // successfully download, we need to delete .zip file
                     String pniZIPFile = pnInbox + "/" + fm.replaceFileName(pnoACKFile, "zip");
                     if (processer.deleteFile(pniZIPFile)) {
                        Util_Logger.debug(logger, "Delete " + pniZIPFile + " successfully");
                        zipsDeleted++;

                     } else {
                        fm.deleteFile(pliACKFile);
                     }

                  } else {

                     // zero length ack file downloaded, so delete it
                     fm.deleteFile(pliACKFile);
                     Util_Logger.debug(logger, "Download " + pnOutbox + "/" + pnoACKFile + " failed, removed");

                  }
               }
               Util_Logger.debug(logger, "Finish FTP process " + pnoACKFile);

            }

         }

         Util_Logger.debug(logger, pnoACKFileList.length + " ACK files have been processed");

      }

      /**********************************************************************
         * Now, we need to identify ACK file type and deal with the local .sent file
         **********************************************************************/
      String[] pliACKFileList = fm.scanDir(plInbox, new BatcherFileFilter(".ack"));

      if ((pliACKFileList != null) && (pliACKFileList.length > 0)) {

         for (int c5 = 0; c5 < pliACKFileList.length; c5++) {

            String pliACKFile = pliACKFileList[c5];
            String pliSENDFile = fm.replaceFileName(pliACKFile, "sent");
            Util_Logger.debug(logger, "Start to validate " + pliACKFile);

            // parse the ACK file
            String ackFileStatus = afp.validateAckFile(plInbox + "/" + pliACKFile);

            if (Constants.POSITIVE_ACK.equals(ackFileStatus)) {
               // positive acknowledge
               // rename .sent to .done
               String pliDONEFile = fm.replaceFileName(pliACKFile, "done");
               fm.renameFile(plInbox, pliSENDFile, pliDONEFile);

               // rename ack file to .ack_done
               String pliACK_DONE = fm.replaceFileName(pliACKFile, "ack_done");
               fm.renameFile(plInbox, pliACKFile, pliACK_DONE);

               // remove resend if it exists in the control file
               String pliRESENDFile = fm.replaceFileName(pliACKFile, "resend");
               ResendTryController.getInstance(Constants.getConstant().getInstanceName()).deleteEntry(pliRESENDFile);

            } else if (Constants.NEGATIVE_ACK.equals(ackFileStatus)) {
               // negative acknowledge
               // rename .sent to .nack
               String pliNACKFile = fm.replaceFileName(pliACKFile, "nack");
               fm.renameFile(plInbox, pliSENDFile, pliNACKFile);

               // rename ack file to .ack_nack
               String pliACK_NACK = fm.replaceFileName(pliACKFile, "ack_nack");
               fm.renameFile(plInbox, pliACKFile, pliACK_NACK);
            } else {
               // Specical 111 code
               // rename .sent to .resend
               String pliRESENDFile = fm.replaceFileName(pliACKFile, "resend");
               fm.renameFile(plInbox, pliSENDFile, pliRESENDFile);
               // delete ack file
               String pliACK_RESEND = fm.replaceFileName(pliACKFile, "ack_resend");
               fm.renameFile(plInbox, pliACKFile, pliACK_RESEND);
            }

            Util_Logger.debug(logger, "Finish validate " + pliACKFile);

         }
      }

      /**********************************************************************
         * Now, we need to process old ack files once NEMMCO has deleted the acks
         * 20-Aug-2004
         **********************************************************************/

      // v 1.4.2
      pliACKFileExistList = fm.scanDir(plInbox, new BatcherFileFilter(new String[] { ".ack_done", ".ack_resend", ".ack_nack" }));
      if (pliACKFileExistList != null) {
         hds.setPliAckFileCount(pliACKFileExistList.length);
      }

      if ((pliACKFileExistList != null) && (pliACKFileExistList.length > 0)) {

         // scan the list of local ack files of type .ack_done, .ack_resend and .ack_nack
         for (int c10 = 0; c10 < pliACKFileExistList.length; c10++) {

            boolean hasMatch = false;
            String pliACKFile = pliACKFileExistList[c10];

            // scan the NEMMCO list to see if there is a match
            for (int c11 = 0; c11 < pnoACKFileList.length; c11++) {

               // if the filenames, at least the start
               if (fm.fileNameCompare(pnoACKFileList[c11].toString(), pliACKFile)) {
                  hasMatch = true;
                  break;
               }

            }

            // if there is no match then we can process the local .ack_.... file
            if (!hasMatch) {

               if (pliACKFile.endsWith(".ack_done")) {
                  // positive acknowledge
                  // move ack file to Done sub dir
                  fm.copyLocalFile(plInbox, plInbox + "/done", pliACKFile);
               } else if (pliACKFile.endsWith(".ack_nack")) {
                  // negative acknowledge
                  // move ack file to nack sub dir
                  fm.copyLocalFile(plInbox, plInbox + "/nack", pliACKFile);
               } else if (pliACKFile.endsWith(".ack_resend")) {
                  // Specical 111 code
                  // delete ack file
                  fm.deleteFile(plInbox + "/" + pliACKFile);
               }

               Util_Logger.debug(logger, "Ack file cleared by NEMMCO " + pliACKFile);

            }

         }
      }

   }

   /**
    * Find out if the file is a b2b file,
    * if so analyse the xml to find the recipient
    * check the stopbox to see if recipient is stopped
    * return true if recipient is stopped
    * 19/05/2004 Mike Muir
    * @param fileName the file name of the potential b2b zip file
    * @return true if the associated participant is stopped and it is a B2B file
    */

   private boolean checkB2B(String fileName) {

      // is b2b file types defined?
      if (Constants.getConstant().getB2BFileTypes().size() > 0) {

         // for each b2b file type
         for (int c6 = 0; c6 < Constants.getConstant().getB2BFileTypes().size(); c6++) {

            // is the file a b2b file based on the start of the filename
            if (fileName.startsWith(Constants.getConstant().getB2BFileTypes().elementAt(c6))) {

               // Open, validate and save some information in zfi
               ZipFileInfo zfi = zfv.validate(plInbox, fileName);

               // success means unzipped XML file
               if (zfi.isProcessSuccess()) {

                  // schema valid XML file
                  if (zfi.isValid()) {

                     // get a list of files in the stop box at the NEMMCO end
                     String[] stopList = processer.listNames(Constants.getConstant().getPNStopboxDir());

                     // check that a list was returned
                     if (stopList != null) {

                        // scan the list
                        for (int c7 = 0; c7 < stopList.length; c7++) {

                           // get file name from /mms/stopbox/coalco_B2bholdinp.stp
                           int index = stopList[c7].lastIndexOf("/");
                           String stopFileName = stopList[c7].substring(index + 1);

                           // if list has filename "participantid*", then that participantid is stopped v1.5.3
                           if (stopFileName.toUpperCase().startsWith((zfi.getToName() + "_").toUpperCase())) {

                              // log and return true to skip the file
                              Util_Logger.debug(logger, "Stopfile for " + zfi.getToName() + " not sending " + fileName);
                              return true;
                           }
                        }
                        //
                     } else {
                        Util_Logger.warning(logger, "Stopbox can not get list of files " + fileName);
                        return true;
                     }
                     //
                  } else {
                     Util_Logger.warning(logger, "Non schema valid send file " + fileName);
                     return false; // v1.5.2 but send it anyway so that it gets -ve acked by the hub
                  }
                  //
               } else {
                  Util_Logger.warning(logger, "Unable to unzip send file " + fileName);
                  return false; // v1.5.2 but send it anyway so that it gets -ve acked by the hub
               }
            }
         }
      }
      return false; // ie: it is ok to process this file
   }

}
