/*
 * Copyright (C) The Apache Software Foundation. All rights reserved.
 *
 * This software is published under the terms of the Apache Software
 * License version 1.1, a copy of which has been included with this
 * distribution in the LICENSE.APL file.
 *
 * Copyright (c) 2001 Cap Gemini Ernst & Young LLC
 * Modification History
 *
 * Date Changed    Changed By         Comments
 * ------------    ----------         --------
 * 14-Jan-2003     Adi Jajuli         Initial
 * 14-Aug-2012     Bob Jones          7984 - Java upgrade from 6 to 7.
 * 10-Feb-2015     Warwick Mills      5252 Metering Batcher fails when duplicate file exists
 */

package com.nemmco.participantbatcher.util;

import java.io.IOException;

import java.text.SimpleDateFormat;

import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import org.apache.log4j.FileAppender;
import org.apache.log4j.Layout;
import org.apache.log4j.spi.ErrorCode;
import org.apache.log4j.spi.LoggingEvent;

/**
 * DailyFileAppender extends FileAppender to use filenames formatted with
 * date/time information. The filename is recomputed every day at midnight.
 *
 * @author Adi Jajuli
 *  version 1.0   - 15-Feb-2001 Created
 * @version 2.0.7 - 10-Feb-2015 Overwrite duplicate files
 */
public class DailyFileAppender extends FileAppender {

   private static SimpleDateFormat dateFormatter = new SimpleDateFormat("yyyy-MM-dd");
   private String currentFileName = null;
   private String fileNameWithoutDate = null;
   private long nextFilenameComputingMillis = System.currentTimeMillis() - 1;

   public DailyFileAppender() {}

   public DailyFileAppender(Layout layout, String filename, boolean append) throws IOException {
      super(layout, filename, append);
   }

   public DailyFileAppender(Layout layout, String filename) throws IOException {
      super(layout, filename);
   }

   public void setFile(String file) {
      fileName = file.trim();
      fileNameWithoutDate = fileName;
   }

   public synchronized void setFile(String fileName, boolean append) throws IOException {

      int dotIdx = fileName.lastIndexOf(".");
      String newName = fileName.substring(0, dotIdx);
      newName += "_";
      newName += dateFormatter.format(new Date());
      newName += fileName.substring(dotIdx);

      if (newName.equals(currentFileName))
         return;

      /* Set up next filename checkpoint */
      GregorianCalendar gc = new GregorianCalendar();
      // next day
      gc.set(Calendar.HOUR_OF_DAY, 0);
      gc.set(Calendar.MINUTE, 0);
      gc.set(Calendar.SECOND, 0);
      gc.set(Calendar.MILLISECOND, 0);
      gc.add(Calendar.DATE, 1);

      nextFilenameComputingMillis = gc.getTime().getTime();

      currentFileName = newName;

      super.setFile(newName, append, bufferedIO, bufferSize);
   }

   protected void subAppend(LoggingEvent event) {
      if (System.currentTimeMillis() >= nextFilenameComputingMillis) {
         try {
            setFile(fileNameWithoutDate, super.fileAppend);
         } catch (IOException e) {
            System.err.println("setFile() call failed.");
            e.printStackTrace();
         }
      }
      super.subAppend(event);
   }

   public void activateOptions() {
      try {
         setFile(fileNameWithoutDate, super.fileAppend);
      } catch (java.io.IOException e) {
         errorHandler.error("setFile(null," + fileAppend + ") call failed.", e, ErrorCode.FILE_OPEN_FAILURE);
      }
   }
}
