/*
 * The Copyright of this work is vested in the National Electricity Market
 * Management Company Limited and the software is issued in confidence
 * for the purpose only for which it is supplied. It must not be
 * reproduced in whole or in part or used for tendering or manufacturing
 * purposes except under an agreement or with the consent in writing of
 * the National Electricity Market Management Company Limited and then
 * only on the condition that this notice is included in any such
 * reproduction.  No information as to the contents or subject matter of
 * this software or any part thereof arising directly or indirectly there
 * from shall be given orally or in writing or communicated in any manner
 * whatsoever to any third party being an individual firm or company or
 * any employee thereof without the prior consent in writing of the National
 * Electricity Market Management Company Limited.
 *
 * Copyright the National Electricity Market Management Company Limited.  2003
 * Mike Muir
 * 28-Aug-2012     Oleg Pevzner       7984 Java upgrade from 6 to 7
 * 06-Mar-2019     Deepak Gupta       DFCT0022009
 */

package com.nemmco.participantbatcher.xml;

import com.nemmco.participantbatcher.util.Constants;
import com.nemmco.participantbatcher.util.MSATSDOMHelper;
import com.nemmco.participantbatcher.util.Util_Logger;

import java.util.Vector;

import org.apache.log4j.Logger;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

/**
 * This class is used to parse the unziped XML file and load some header datea
 * into ZipFileInfo.
 *
 * author Kevin Jin
 * version 1 (18/09/2003) -- Creation
 * version 1.3 04-Oct-2004 - log4j
 * version 1.3.1 16-Nov-2004 When eXerces throws exception use as the -ve ack message
 * version 1.4.3 22-Mar-2006 Defect 3771 Null Pointer with null security context in header of zip file
 * version 1.4.4 6-May-2006 Fixed javaDoc and coding standards
 * version 1.5.3 - 10-Jul-2009 5796 Get schemaVersion from zip and set it in ZipFileInfo object.
 * version 2.0.1 15-Nov-2010 Peer review work
 * version 2.0.2 - 19-Nov-2010 Peer review work
 */

public class ZipFileParser {

   private static Logger logger = Logger.getLogger(ZipFileParser.class);

   /**
    * Constructor
    */
   public ZipFileParser() {}

   /** Do the parsing of the file.
    *
    * @param fileName the file name to parse
    * @param zfi the ZipFileInfo to update
    */
   public void parser(String fileName, ZipFileInfo zfi) {

      BatcherDOMParser parser = new BatcherDOMParser();
      Document document = null;
      try {

         if (Constants.getConstant().isValidateSchema()) {
            document = parser.createDocument(fileName, true);
         } else {
            document = parser.createDocument(fileName, false);
         }

         // if there is no exception occured up to this point,
         // we should be able to say that everything is OK
         if (parser.getErrorMessage().trim().length() > 0) {
            zfi.setIsValid(false);
            zfi.setErrorMessage(parser.getErrorMessage());
         } else {
            zfi.setIsValid(true);
         }

      } catch (Exception e) {

         // Something wrong when parsing the XML file
         zfi.setIsValid(false);
         zfi.setErrorMessage(e.getMessage()); //parser.getErrorMessage() );
         Util_Logger.warning(logger, "Exception occured when parsing " + fileName, e);
      }

      if (document != null) {

         String zipSchemaNamespace = MSATSDOMHelper.getAttribute(document, "ase:aseXML", "xmlns:ase");
         String[] zipSchemaNamespaceComponents = zipSchemaNamespace.split(":");
         String schemaVersion = zipSchemaNamespaceComponents[2];
         String fromName = MSATSDOMHelper.getElement(document, "From");
         String fromDesc = MSATSDOMHelper.getAttribute(document, "From", "description");
         String toName = MSATSDOMHelper.getElement(document, "To");
         String toDesc = MSATSDOMHelper.getAttribute(document, "To", "description");
         String messageId = MSATSDOMHelper.getElement(document, "MessageID");
         String transGroup = MSATSDOMHelper.getElement(document, "TransactionGroup");
         String priority = MSATSDOMHelper.getElement(document, "Priority");
         String securityContext = MSATSDOMHelper.getElement(document, "SecurityContext");

			zfi.setFileEligibleToBeProcessed(isFileEligible(document));

         zfi.setSchemaVersion(schemaVersion);
         zfi.setPriority(priority);
         zfi.setSecurityContext(securityContext);
         zfi.setTransGroup(transGroup);
         zfi.setInitMsgId(messageId);
         zfi.setFromName(fromName);
         zfi.setFromDesc(fromDesc);
         zfi.setToName(toName);
         zfi.setToDesc(toDesc);
      }

   }

	private boolean isFileEligible(Document document) {
	   String schemaVersion;
		Element transaction;
		Element transactionType;
		String transactionTypeName;
	   boolean txTypeAllowed;
	   boolean schemaAllowed = true;
		boolean fileEligible = true;

	   Element root = MSATSDOMHelper.getElementNode(document, "ase:aseXML");

	   // determine schema version
	   String xmlns = MSATSDOMHelper.getAttribute(document, "ase:aseXML", "xmlns:ase");
	   if (xmlns != null) {
	      schemaVersion = xmlns.substring(xmlns.lastIndexOf(":") + 1);
	      schemaAllowed = isSchemaAllowed(schemaVersion);
	   }
		fileEligible = schemaAllowed;

	   // If schema version is in the acceptSchemas list, then only check for tx types.
		if (fileEligible) {
		   Vector<Element> transactionList = MSATSDOMHelper.getElementNodes(document, "ase:aseXML/Transactions/Transaction");

		   if (!transactionList.isEmpty()) {
		      //process each transaction one at a time
		      for (int i = 0; i < transactionList.size(); i++) {
		         transaction = transactionList.get(i);

		         //retrieve the transaction type
		         transactionType = MSATSDOMHelper.getFirstChildElementNode(transaction);
		         transactionTypeName = transactionType.getNodeName();

		         txTypeAllowed = isTxTypeAllowed(transactionTypeName);

	            fileEligible = txTypeAllowed;
		      }
		   }
		}
		return fileEligible;
	}

	private boolean isTxTypeAllowed(String txTypeName) {
		boolean txTypeAllowed = true;
	   Vector<String> acceptTxTypes = Constants.getConstant().getAcceptTxTypes();
	   Vector<String> rejectTxTypes = Constants.getConstant().getRejectTxTypes();

		if (null != acceptTxTypes && !acceptTxTypes.isEmpty()) {
			txTypeAllowed = false;
		   for(String allowedTxType : acceptTxTypes) {
		      if (txTypeName.equalsIgnoreCase(allowedTxType)) {
		         txTypeAllowed = true;
		         break;
		      }
		   }
		} else if (null != rejectTxTypes && !rejectTxTypes.isEmpty()) {
			for(String rejectTxType : rejectTxTypes) {
				if (txTypeName.equalsIgnoreCase(rejectTxType)) {
					txTypeAllowed = false;
					break;
				}
			}
		}
		return txTypeAllowed;
	}

	private boolean isSchemaAllowed(String schemaName) {
		boolean schemaAllowed = true;
	   Vector<String> acceptSchemas = Constants.getConstant().getAcceptSchemas();
	   Vector<String> rejectSchemas = Constants.getConstant().getRejectSchemas();

		if (null != acceptSchemas && !acceptSchemas.isEmpty()) {
			schemaAllowed = false;
		   for(String allowedSchema : acceptSchemas) {
		      if (schemaName.equalsIgnoreCase(allowedSchema)) {
		         schemaAllowed = true;
		         break;
		      }
		   }
		} else if (null != rejectSchemas && !rejectSchemas.isEmpty()) {
		   for(String rejectSchema : rejectSchemas) {
		      if (schemaName.equalsIgnoreCase(rejectSchema)) {
		         schemaAllowed = false;
		         break;
		      }
		   }
		}
		return schemaAllowed;
	}
}
